﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using DroidExplorer.Core.UI;
using DroidExplorer.UI;
using DroidExplorer.Core.Plugins;
using DroidExplorer.Core;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Globalization;

namespace DroidExplorer.Plugins.UI {
	/// <summary>
	/// 
	/// </summary>
	public partial class ShellConsole : PluginForm, IShellProcess {
		private delegate int GetTextLengthDelegate( );
		private delegate void SetSelectionStartDelegate( int start );
		private delegate void SetStatusLabelTextDelegate( ToolStripStatusLabel label, string text );
		private delegate IntPtr GetControlHandleDelegate( Control ctrl );
		private delegate int GetLineFromCharIndexDelegate( int index );
		private delegate int GetSelectionStartDelegate( );
		private delegate int GetShellLinesLengthDelegate( );
		private delegate string[] GetShellLinesDelegate( );
		private delegate bool GenericBooleanDelegate( );

		#region Winapi imports
		[DllImport ( "user32.dll" )]
		private static extern int CreateCaret( IntPtr hwnd, IntPtr hbm, int cx, int cy );

		[DllImport ( "user32.dll" )]
		private static extern int DestroyCaret( );

		[DllImport ( "user32.dll" )]
		private static extern int SetCaretPos( int x, int y );

		[DllImport ( "user32.dll" )]
		private static extern int ShowCaret( IntPtr hwnd );

		[DllImport ( "user32.dll" )]
		private static extern int HideCaret( IntPtr hwnd );

		[DllImport ( "User32.Dll" )]
		private static extern int SendMessage( IntPtr hWnd, int Msg, int wParam, int lParam );
		private const int EM_LINEINDEX = 0xBB;
		#endregion

		private ConsoleColor _caretColor = ConsoleColor.Green;
		private Size _caretSize = new Size ( 8, 16 );

		/// <summary>
		/// Initializes a new instance of the <see cref="ShellConsole"/> class.
		/// </summary>
		public ShellConsole( ) {
			InitializeComponent ( );
			CaretSize = new Size ( 8, 16 );
			CaretColor = ConsoleColor.Green;
			this.InvalidateCaret ( );
			this.ShellType = "#";
			this.AltShellType = "$";
			Commands = new List<string> ( );
			CommandListIndex = -1;
			this.StickyWindow = new StickyWindow ( this );
			ConsoleOutput = new ConsoleWriter ( ref this.shell );
			ConsoleError = new ConsoleWriter ( ref this.shell );
			this.shell.KeyUp += new KeyEventHandler ( shell_KeyUp );
			this.shell.KeyDown += new KeyEventHandler ( shell_KeyDown );
			this.shell.KeyPress += new KeyPressEventHandler ( shell_KeyPress );
			this.shell.SelectionChanged += new EventHandler ( shell_SelectionChanged );
			this.shell.TextChanged += new EventHandler ( shell_TextChanged );
			this.shell.MouseDown += new MouseEventHandler ( shell_MouseDown );
			this.shell.MouseUp += new MouseEventHandler ( shell_MouseUp );
			this.shell.GotFocus += new EventHandler ( shell_GotFocus );
			this.shell.HandleCreated += new EventHandler ( shell_HandleCreated );
			Run ( String.Empty );

		}


		#region Overrides

		/// <summary>
		/// Raises the <see cref="E:System.Windows.Forms.Control.Validating"/> event.
		/// </summary>
		/// <param name="e">A <see cref="T:System.ComponentModel.CancelEventArgs"/> that contains the event data.</param>
		protected override void OnValidating( CancelEventArgs e ) {
			ShowHideCaret ( );
			base.OnValidating ( e );
		}

		/// <summary>
		/// </summary>
		/// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
		protected override void OnResize( EventArgs e ) {
			ShowHideCaret ( );
			base.OnResize ( e );
		}

		/// <summary>
		/// </summary>
		/// <param name="e">A <see cref="T:System.Windows.Forms.PaintEventArgs"/> that contains the event data.</param>
		protected override void OnPaint( PaintEventArgs e ) {
			base.OnPaint ( e );
			ShowHideCaret ( );
		}

		/// <summary>
		/// Raises the <see cref="E:System.Windows.Forms.Control.Enter"/> event.
		/// </summary>
		/// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
		protected override void OnEnter( EventArgs e ) {
			ShowHideCaret ( );
			base.OnEnter ( e );
		}

		/// <summary>
		/// Raises the <see cref="E:System.Windows.Forms.Form.Activated"/> event.
		/// </summary>
		/// <param name="e">An <see cref="T:System.EventArgs"/> that contains the event data.</param>
		protected override void OnActivated( EventArgs e ) {
			ShowHideCaret ( );
			base.OnActivated ( e );
		}

		/// <summary>
		/// Raises the <see cref="E:System.Windows.Forms.Form.Closing"/> event.
		/// </summary>
		/// <param name="e">A <see cref="T:System.ComponentModel.CancelEventArgs"/> that contains the event data.</param>
		protected override void OnClosing( CancelEventArgs e ) {
			if ( IsRunning ) {
				if ( !this.BaseProcess.HasExited ) {
					BaseProcess.Kill ( );
				}
				BaseProcess = null;
			}
			base.OnClosing ( e );
		}

		#endregion

		#region Properties (Private & Public)
		private List<string> Commands { get; set; }
		private int CommandListIndex { get; set; }
		private string CurrentCommand { get; set; }
		private bool IsInEditableArea { get; set; }
		private bool OutputingData { get; set; }
		private string ShellType { get; set; }
		private string AltShellType { get; set; }
		private Bitmap Caret { get; set; }
		private bool IsRunning { get { return this.BaseProcess != null; } }
		private int LastEditableCharacterPosition { get; set; }
		private ConsoleWriter ConsoleOutput { get; set; }
		private ConsoleWriter ConsoleError { get; set; }

		/// <summary>
		/// Gets or sets the current location.
		/// </summary>
		/// <value>The current location.</value>
		public Point CurrentLocation { get; set; }
		/// <summary>
		/// Gets or sets the color of the caret.
		/// </summary>
		/// <value>The color of the caret.</value>
		public ConsoleColor CaretColor { get { return _caretColor; } set { _caretColor = value; InvalidateCaret ( ); } }
		/// <summary>
		/// Gets or sets the size of the caret.
		/// </summary>
		/// <value>The size of the caret.</value>
		public Size CaretSize { get { return _caretSize; } set { _caretSize = value; InvalidateCaret ( ); } }

		#endregion


		#region Shell RTB Event Handlers


		/// <summary>
		/// Handles the HandleCreated event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void shell_HandleCreated( object sender, EventArgs e ) {

		}


		/// <summary>
		/// Handles the TextChanged event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void shell_TextChanged( object sender, EventArgs e ) {
			ShowHideCaret ( );
		}

		/// <summary>
		/// Handles the MouseDown event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
		void shell_MouseDown( object sender, MouseEventArgs e ) {
			this.shell.SelectionStart = this.shell.TextLength;
			ShowHideCaret ( );
		}

		/// <summary>
		/// Handles the MouseUp event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
		void shell_MouseUp( object sender, MouseEventArgs e ) {
			this.shell.SelectionStart = this.shell.TextLength;
			ShowHideCaret ( );
		}


		/// <summary>
		/// Handles the SelectionChanged event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void shell_SelectionChanged( object sender, EventArgs e ) {
			IsInEditableArea = shell.SelectionStart >= LastEditableCharacterPosition;
			GetLocation ( );
			ShowHideCaret ( );
		}

		/// <summary>
		/// Handles the GotFocus event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void shell_GotFocus( object sender, EventArgs e ) {

		}

		/// <summary>
		/// Handles the KeyPress event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyPressEventArgs"/> instance containing the event data.</param>
		void shell_KeyPress( object sender, KeyPressEventArgs e ) {

			GetLocation ( );

			if ( e.KeyChar == 9 ) {
				e.Handled = true;
			}
		}

		/// <summary>
		/// Handles the KeyDown event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyEventArgs"/> instance containing the event data.</param>
		void shell_KeyDown( object sender, KeyEventArgs e ) {
			if ( ( e.KeyCode == Keys.C || e.KeyCode == Keys.X ) && e.Control ) {
				Console.WriteLine ( "CTRL+{0}", e.KeyCode.ToString ( ) );
				if ( InputStream != null ) {
					if ( BaseProcess != null && !BaseProcess.HasExited ) {
						try { BaseProcess.Kill ( ); } catch { }
					}
				}
				e.Handled = true;
				return;
			}


			if ( !IsInEditableArea ) {
				e.Handled = true;
				return;
			}

			if ( ( e.KeyCode == Keys.Back || e.KeyCode == Keys.Left ) && CurrentLocation.X <= this.ShellType.Length + 2 ) {
				e.Handled = true;
			}

			if ( e.KeyCode == Keys.Down || e.KeyCode == Keys.Up ) {
				e.Handled = true;
			}

			if ( e.KeyCode == Keys.Return || e.KeyCode == Keys.Enter ) {
				LastEditableCharacterPosition = this.shell.TextLength;
				this.shell.SelectionStart = this.shell.TextLength;
				e.Handled = true;
				SendCommand ( );
			}

			if ( e.KeyCode == Keys.Tab ) {
				e.Handled = true;
			}


			ShowHideCaret ( );
		}



		/// <summary>
		/// Handles the KeyUp event of the shell control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyEventArgs"/> instance containing the event data.</param>
		void shell_KeyUp( object sender, KeyEventArgs e ) {


		}

		#endregion

		private void GetLocation( ) {
			if ( shell == null || shell.IsDisposed ) {
				return;
			}
			int currentRow = 1;
			int currentCol = 1;
			int rowStartIndex = 0;
			IntPtr shellHandle = IntPtr.Zero;
			int selStart = 0;

			if ( this.InvokeRequired ) {
				selStart = (int)this.Invoke ( new GetSelectionStartDelegate ( GetShellSelectionStart ) );
				currentRow = (int)this.Invoke ( new GetLineFromCharIndexDelegate ( GetLineFromCharIndex ), selStart ) + 1;
				int linesLength = (int)this.Invoke ( new GetShellLinesLengthDelegate ( GetShellLinesLength ) );
				if ( currentRow > linesLength ) {
					currentRow = linesLength;
				}
				shellHandle = (IntPtr)this.Invoke ( new GetControlHandleDelegate ( GetControlHandle ), shell );
			} else {
				selStart = shell.SelectionStart;
				currentRow = shell.GetLineFromCharIndex ( selStart ) + 1;
				if ( currentRow > shell.Lines.Length ) {
					currentRow = shell.Lines.Length;
				}
				shellHandle = shell.Handle;
			}

			rowStartIndex = SendMessage ( shellHandle, EM_LINEINDEX, -1, 0 );
			currentCol = selStart - rowStartIndex + 1;

			this.CurrentLocation = new Point ( currentCol, currentRow );

			if ( !OutputingData ) {
				Point position = new Point ( this.CurrentLocation.X, this.CurrentLocation.Y );
				if ( this.InvokeRequired ) {
					string[] lines = (string[])this.Invoke ( new GetShellLinesDelegate ( GetShellLines ) );
					if ( lines[currentRow - 1].StartsWith ( this.ShellType ) ) {
						position.Offset ( -( this.ShellType.Length + 1 ), 0 );
					} else if ( lines[currentRow - 1].StartsWith ( this.AltShellType ) ) {
						position.Offset ( -( this.AltShellType.Length + 1 ), 0 );
					}
					this.Invoke ( new SetStatusLabelTextDelegate ( SetStatusLabelText ), rowLabel, string.Format ( CultureInfo.InvariantCulture, "Row: {0}", position.Y ) );
					this.Invoke ( new SetStatusLabelTextDelegate ( SetStatusLabelText ), columnLabel, string.Format ( CultureInfo.InvariantCulture, "Column: {0}", position.X ) );
				} else {
					if ( this.shell.Lines[currentRow - 1].StartsWith ( this.ShellType ) ) {
						position.Offset ( -( this.ShellType.Length + 1 ), 0 );
					} else if ( this.shell.Lines[currentRow - 1].StartsWith ( this.AltShellType ) ) {
						position.Offset ( -( this.AltShellType.Length + 1 ), 0 );
					}
					SetStatusLabelText ( rowLabel, string.Format ( CultureInfo.InvariantCulture, "Row: {0}", position.Y ) );
					SetStatusLabelText ( columnLabel, string.Format ( CultureInfo.InvariantCulture, "Column: {0}", position.X ) );
				}
			}
		}

		/// <summary>
		/// Gets the shell lines.
		/// </summary>
		/// <returns></returns>
		private string[] GetShellLines( ) {
			return this.shell.Lines;
		}

		/// <summary>
		/// Gets the length of the shell lines.
		/// </summary>
		/// <returns></returns>
		private int GetShellLinesLength( ) {
			return this.shell.Lines.Length;
		}

		/// <summary>
		/// Gets the shell selection start.
		/// </summary>
		/// <returns></returns>
		private int GetShellSelectionStart( ) {
			return this.shell.SelectionStart;
		}

		/// <summary>
		/// Gets the index of the line from char.
		/// </summary>
		/// <param name="index">The index.</param>
		/// <returns></returns>
		private int GetLineFromCharIndex( int index ) {
			return this.shell.GetLineFromCharIndex ( index );
		}

		/// <summary>
		/// Sets the status label text.
		/// </summary>
		/// <param name="label">The label.</param>
		/// <param name="text">The text.</param>
		private void SetStatusLabelText( ToolStripStatusLabel label, string text ) {
			label.Text = text;
		}

		/// <summary>
		/// Gets the control handle.
		/// </summary>
		/// <param name="ctrl">The CTRL.</param>
		/// <returns></returns>
		private IntPtr GetControlHandle( Control ctrl ) {
			if ( ctrl != null && !ctrl.IsDisposed ) {
				return ctrl.Handle;
			} else {
				return IntPtr.Zero;
			}
		}

		/// <summary>
		/// Invalidates the caret.
		/// </summary>
		private void InvalidateCaret( ) {
			if ( Caret != null ) {
				Caret.Dispose ( );
				Caret = null;
			}

			Caret = new Bitmap ( CaretSize.Width, CaretSize.Height );
			using ( Graphics g = Graphics.FromImage ( Caret ) ) {
				using ( SolidBrush b = new SolidBrush ( ConsoleCharacter.ConsoleColorToColor ( this.CaretColor ) ) ) {
					g.FillRectangle ( b, new Rectangle ( 0, 0, Caret.Width, Caret.Height ) );
				}
			}
			ShowHideCaret ( );
		}

		/// <summary>
		/// Shows the hide caret.
		/// </summary>
		private void ShowHideCaret( ) {
			if ( Caret == null ) {
				InvalidateCaret ( );
			}
			IntPtr handle = IntPtr.Zero;
			if ( this.InvokeRequired ) {
				handle = (IntPtr)this.Invoke ( new GetControlHandleDelegate ( GetControlHandle ), this.shell );
			} else {
				handle = GetControlHandle ( this.shell );
			}

			HideCaret ( handle );
			CreateCaret ( handle, this.Caret.GetHbitmap ( ), CaretSize.Width, CaretSize.Height );
			ShowCaret ( handle );
		}

		private void WritePrompt( ) {
			ConsoleOutput.Write ( "{0} ", this.ShellType );
			if ( this.InvokeRequired ) {
				LastEditableCharacterPosition = (int)this.Invoke ( new GetTextLengthDelegate ( GetTextLength ) );
				this.Invoke ( new SetSelectionStartDelegate ( SetSelectionStart ), LastEditableCharacterPosition );
				this.Invoke ( new GenericBooleanDelegate ( shell.Focus ) );
			} else {
				LastEditableCharacterPosition = this.GetTextLength ( );
				SetSelectionStart ( LastEditableCharacterPosition );
				shell.Focus ( );
			}
			ShowHideCaret ( );
			GetLocation ( );
		}

		/// <summary>
		/// Writes the command.
		/// </summary>
		/// <param name="CurrentCommand">The current command.</param>
		private void WriteCommand( string CurrentCommand ) {
			this.shell.SelectionStart = this.shell.TextLength;
			ConsoleOutput.Write ( CurrentCommand );
			this.shell.SelectionStart = this.shell.TextLength;
		}


		/// <summary>
		/// Gets the last line.
		/// </summary>
		/// <returns></returns>
		private string GetLastLine( ) {
			if ( this.shell.Lines.Length > 0 ) {
				string line = this.shell.Lines[this.shell.Lines.Length - 1];
				if ( line.StartsWith ( this.ShellType ) ) {
					ShellType = line.Substring ( 0, 1 );
					this.LogDebug ( "Set ShellType: {0}", this.ShellType );
					line = line.Substring ( this.ShellType.Length + 1 ).Trim ( );
				}
				return line;
			} else {
				return string.Empty;
			}
		}

		/// <summary>
		/// Sends the command.
		/// </summary>
		private void SendCommand( ) {
			if ( this.shell.Lines.Length > 0 ) {
				string line = GetLastLine ( );
				this.LogDebug ( "SendCommand: {0}", line );
				SendCommand ( line );
			}
		}

		/// <summary>
		/// Sends the command.
		/// </summary>
		/// <param name="command">The command.</param>
		private void SendCommand( string command ) {
			this.LogDebug ( "Send Command: {0}", command );
			ClearLastLine ( );
			ConsoleOutput.WriteLine ( string.Empty );
			Commands.Add ( command );
			switch ( command ) {
				case "exit":
					this.Close ( );
					break;
				case "clear":
					this.shell.Text = string.Empty;
					WritePrompt ( );
					break;
				default:
					if ( IsRunning ) {
						InputStream.WriteLine ( command );
					} else {
						CommandRunner.Instance.LaunchRedirectedShellWindow ( CommandRunner.Instance.DefaultDevice, command, this );
					}
					break;

			}
		}

		/// <summary>
		/// Clears the last line.
		/// </summary>
		private void ClearLastLine( ) {
			if ( this.shell.Lines.Length > 0 ) {
				this.shell.Lines[this.shell.Lines.Length - 1] = string.Empty;
			}
		}

		/// <summary>
		/// Runs the specified command.
		/// </summary>
		/// <param name="command">The command.</param>
		public void Run( string command ) {
			this.LogDebug ( "Run: {0}", command );
			CommandRunner.Instance.LaunchRedirectedShellWindow ( CommandRunner.Instance.DefaultDevice, command, this );
		}

		/// <summary>
		/// Gets the length of the text.
		/// </summary>
		/// <returns></returns>
		private int GetTextLength( ) {
			if ( this.shell != null && !this.shell.IsDisposed ) {
				return this.shell.TextLength;
			} else {
				return 0;
			}
		}

		/// <summary>
		/// Sets the selection start.
		/// </summary>
		/// <param name="start">The start.</param>
		private void SetSelectionStart( int start ) {
			if ( this.shell != null && !this.shell.IsDisposed ) {
				this.shell.SelectionStart = start;
			}
		}

		#region IShellProcess Members

		/// <summary>
		/// Gets the output stream.
		/// </summary>
		/// <value>The output stream.</value>
		public System.IO.StreamReader OutputStream { get; private set; }

		/// <summary>
		/// Gets the error stream.
		/// </summary>
		/// <value>The error stream.</value>
		public System.IO.StreamReader ErrorStream { get; private set; }

		/// <summary>
		/// Gets the input stream.
		/// </summary>
		/// <value>The input stream.</value>
		public System.IO.StreamWriter InputStream { get; private set; }

		/// <summary>
		/// Gets the base process.
		/// </summary>
		/// <value>The base process.</value>
		public System.Diagnostics.Process BaseProcess { get; private set; }

		/// <summary>
		/// Sets the base process.
		/// </summary>
		/// <param name="process">The process.</param>
		public void SetBaseProcess( System.Diagnostics.Process process ) {
			if ( process != null ) {
				BaseProcess = process;
			} else {
				WritePrompt ( );
			}
		}

		/// <summary>
		/// Sets the output stream.
		/// </summary>
		/// <param name="reader">The reader.</param>
		public void SetOutputStream( System.IO.StreamReader reader ) {
			OutputStream = reader;
			new Thread ( new ThreadStart ( delegate {
				while ( IsRunning ) {

					while ( !OutputStream.EndOfStream ) {
						int peekedCount = 0;
						if ( OutputStream.Peek ( ) == '/' ) {
							while ( ( (char)OutputStream.Peek ( ) != 35 && (char)OutputStream.Peek ( ) != 36 ) ) {
								OutputStream.Read ( );
								++peekedCount;
							}
						}
						char nextChar = (char)OutputStream.Peek ( );
						if ( nextChar == 35 || nextChar == 36 ) {
							this.ShellType = new String ( new char[] { nextChar } );
							WritePrompt ( );
							OutputingData = false;
							OutputStream.ReadLine ( );
						} else {
							string line = OutputStream.ReadLine ( );
							OutputingData = true;
							if ( !string.IsNullOrEmpty ( line ) && !line.StartsWith ( this.ShellType ) && !line.StartsWith ( this.AltShellType ) ) {
								ConsoleOutput.WriteLine ( line );
							}
						}
					}
				}
			} ) ).Start ( );
		}

		/// <summary>
		/// Sets the error stream.
		/// </summary>
		/// <param name="reader">The reader.</param>
		public void SetErrorStream( System.IO.StreamReader reader ) {
			ErrorStream = reader;
		}

		/// <summary>
		/// Sets the input stream.
		/// </summary>
		/// <param name="writer">The writer.</param>
		public void SetInputStream( System.IO.StreamWriter writer ) {
			InputStream = writer;
		}

		#endregion
	}
}
