using System;
using System.Linq;
using System.Collections.Generic;
using System.Windows.Forms;
using DroidExplorer.UI;
using DroidExplorer.Core;
using System.IO;
using System.Drawing;
using DroidExplorer.Core.Plugins;
using DroidExplorer.Core.UI;
using DroidExplorer.Tools;
using DroidExplorer.Configuration;
using log4net.Config;
using DroidExplorer.Controls;
using DroidExplorer.Configuration.Net;
using System.Threading;

namespace DroidExplorer {
	/// <summary>
	/// 
	/// </summary>
	static class Program {
		private static PackageManagerForm _packageManager = null;
		private static List<IPlugin> _loadedPlugins = null;

		/// <summary>
		/// Initializes the <see cref="Program"/> class.
		/// </summary>
		static Program ( ) {

		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main ( string[] arguments ) {
			try {
				using ( var stream = typeof ( Program ).Assembly.GetManifestResourceStream ( DroidExplorer.Resources.Strings.Log4NetDroidExplorer ) ) {
					XmlConfigurator.Configure ( stream );
				}

				var args = new Arguments ( arguments );
				Application.EnableVisualStyles ( );
				Application.SetCompatibleTextRenderingDefault ( false );
				Application.SetUnhandledExceptionMode ( UnhandledExceptionMode.CatchException );
				Application.ThreadException +=
					( sender, e ) => DroidExplorer.Core.Logger.LogError ( typeof (Program), e.Exception.Message, e.Exception );

				// show the splash dialog
				Program.SplashManager.ShowSplashDialog ( 13 );

				// load settings
				Program.SplashManager.SplashDialog.SetStepText ( "Loading Settings..." );
				CommandRunner.Settings = Settings.Instance;
				Settings.Instance.SystemSettings = RegistrySettings.Instance;
				Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );
		
				// cleanup adb? should this be done?
				Application.ApplicationExit += delegate ( object sender, EventArgs e ) {
					//CommandRunner.Instance.AdbProcessCleanUp ( );
				};

				// locate sdk tools
				Program.SplashManager.SplashDialog.SetStepText ( DroidExplorer.Resources.Strings.SplashLocatingSdk );
				var sdkPath = Settings.Instance.SystemSettings.SdkPath;

				if ( string.IsNullOrEmpty ( sdkPath ) ) {
					throw new ApplicationException ( "Unable to locate the android SDK tools, please reinstall the application." );
				}
				Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );

				// verify the sdk tools
				Program.SplashManager.SplashDialog.SetStepText ( "Verifying SDK Tools" );
				if ( !CommandRunner.Instance.VerifyAndroidSdkTools ( ) ) {
					Program.SplashManager.CloseSplashDialog ( );
					throw new ApplicationException ( DroidExplorer.Resources.Strings.SdkPathNotFoundMessage );
				}
				Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );

				// set the sdk path
				CommandRunner.Instance.SdkPath = sdkPath;
				Program.SplashManager.SplashDialog.SetStepText ( DroidExplorer.Resources.Strings.SplashStartAdb);
				// start the adb server
				CommandRunner.Instance.StartServer ( );
				Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );

				// get the attached devices
				Program.SplashManager.SplashDialog.SetStepText ( DroidExplorer.Resources.Strings.SplashGetDevices );
				// BUG: 15695
				// TODO: Fix this:
				// this is a pain in the ass bug. for some reason it will not get the devices the first time.
				// the shell call actually returns that there are no devices attached.
				var devices = CommandRunner.Instance.GetDevices ( );
				Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );

				// are we attaching to a specific device?
				if ( args.Contains ( "d", "device" ) ) {
					CommandRunner.Instance.DefaultDevice = args[ "d", "device" ];
					Logger.LogDebug ( DroidExplorer.Resources.Strings.SplashConnectingDevice, CommandRunner.Instance.DefaultDevice );
					Application.Run ( new MainForm ( ) );
				} else {
					// get the attached devices
					String selectedDevice = null;
					if ( devices.Count != 1 ) {
						// we have 0 or > 1 so we need to display the "selection dialog"
						var dsf = new GenericDeviceSelectionForm ( );
						Program.SplashManager.SplashDialog.HideExt ( );
						if ( dsf.ShowDialog ( null ) == DialogResult.OK ) {
							CommandRunner.Instance.DefaultDevice = dsf.SelectedDevice;
							selectedDevice = dsf.SelectedDevice;
						} else {
							Program.SplashManager.CloseSplashDialog ( );
							return;
						}
						Program.SplashManager.SplashDialog.AsyncShow ( );
					} else {
						selectedDevice = devices[0];
						CommandRunner.Instance.DefaultDevice = selectedDevice;
					}

					// minor hackory for getting the device for cloud stats
					var d = new Core.Adb.Device(selectedDevice,Core.Adb.DeviceState.Online);
					// get the device properties for registration
					foreach ( var i in CommandRunner.Instance.GetProperties ( selectedDevice ) ) {
						d.Properties.Add ( i.Key, i.Value );
					}
					if ( !Settings.Instance.SystemSettings.RecordDeviceInformationToCloud ) {
						Program.SplashManager.SplashDialog.SetStepText (
							String.Format ( DroidExplorer.Resources.Strings.SplashRegisterDevice,
								KnownDeviceManager.Instance.GetDeviceFriendlyName ( selectedDevice )
							)
						);
						// register the device
						CloudStatistics.Instance.RegisterDevice ( d );
						Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );
					} else {
						// we have to increment but nothing else here
						Program.SplashManager.SplashDialog.IncrementLoadStep ( 1 );
					}
					// connect to device
					Program.SplashManager.SplashDialog.SetStepText ( string.Format ( DroidExplorer.Resources.Strings.SplashConnectingDevice, KnownDeviceManager.Instance.GetDeviceFriendlyName ( CommandRunner.Instance.DefaultDevice ) ) );
					Application.Run ( new MainForm ( ) );
				}
			} catch ( Exception ex ) {
				var result = TaskDialog.ShowCommandBox ( DroidExplorer.Resources.Strings.ErrorUncaughtTitle, ex.Message, 
					DroidExplorer.Resources.Strings.ErrorUncaughtMessage, ex.ToString ( ), 
					string.Empty, string.Empty, DroidExplorer.Resources.Strings.ErrorUncaughtCommands, 
					false, SysIcons.Error, SysIcons.Error );
				switch ( TaskDialog.CommandButtonResult ) {
					case 1:
						// continue
						break;
					case 2:
						// report bug
						CommandRunner.Instance.LaunchProcessWindow ( DroidExplorer.Resources.Strings.IssueTrackerCreateUrl, string.Empty, true );
						break;
					default: // 0
						Application.Restart ( );
						break;
				}
			}
		}

		/// <summary>
		/// Gets the splash manager.
		/// </summary>
		/// <value>The splash manager.</value>
		public static SplashManager SplashManager {
			get {
				return SplashManager.Instance;
			}
		}


		/// <summary>
		/// Gets the loaded plugins.
		/// </summary>
		/// <value>The loaded plugins.</value>
		public static List<IPlugin> LoadedPlugins {
			get { return _loadedPlugins ?? (_loadedPlugins = new List<IPlugin> ( )); }
		}

		/// <summary>
		/// Gets the package manager.
		/// </summary>
		/// <value>The package manager.</value>
		public static PackageManagerForm PackageManager {
			get {
				if ( _packageManager == null || _packageManager.IsDisposed ) {
					_packageManager = new PackageManagerForm ( );
				}
				return _packageManager;
			}
		}

		/// <summary>
		/// Gets the system icons.
		/// </summary>
		/// <value>The system icons.</value>
		public static Dictionary<string, int> SystemIcons {
			get {
				return SystemImageListHost.Instance.SystemIcons;
			}
		}
		/// <summary>
		/// Gets the local path.
		/// </summary>
		/// <value>The local path.</value>
		public static string LocalPath {
			get {
				return Path.GetDirectoryName ( typeof ( Program ).Assembly.Location );
			}
		}
	}
}